# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.ssh
:synopsis: A Package for SSH management and communication.
:author: Amr Ali <amr@databracket.com>

.. rst-class:: html-toggle

Backend Connection Management
-----------------------------
.. automodule:: bastio.ssh.client

.. rst-class:: html-toggle

Action Processor
----------------
.. automodule:: bastio.ssh.api

.. rst-class:: html-toggle

Protocol Parsing And Logic Routing
----------------------------------
.. automodule:: bastio.ssh.protocol

.. rst-class:: html-toggle

Cryptographic Utilities
-----------------------
.. automodule:: bastio.ssh.crypto
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

